<?php
/**
 * @file
 * Cache class wrapper for non-transactional backend.
 *
 * The class works in two modes: safe or non-safe.
 *
 * Both modes buffers cache operations during transactions, in order to keep
 * cache in sync with database contents.
 *
 * In safe-mode however, cache_set() is silenty ignored/discarded during
 * transactions.
 *
 * Safe mode is needed for fully eliminating the race conditions that cause
 * the cache to come out of sync with the DB, if the REPEATABLE-READ isolation
 * level is used.
 *
 * If using the READ-COMMITTED isolation level, safe mode should not be enabled.
 *
 * @code
 *  $conf['consistent_cache_default_safe'] =  FALSE;
 * @endcode
 */

/**
 * Constants for buffer array indices.
 *
 * Let's use constants instead of strings, for reasons of compile-time
 * optimization.
 */
const CC_DEPTH = 0;
const CC_OPERATION = 1;
const CC_CID = 2;
const CC_DATA_WILDCARD = 3;
const CC_EXPIRE = 4;

/**
 * Helper function for getting bin specific variables with fallback.
 *
 * @param string $prefix
 *   The variable prefix, e.g. 'consistent_cache'.
 * @param string $bin
 *   The name of the bin, e.g. 'cache'.
 * @param string $name
 *   The name of the variable, e.g. 'class'.
 * @param string $default
 *   The default value if none found, e.g. 'DrupalDatabaseCache'.
 *
 * @return mixed
 *   The value of the variable.
 */
function cache_consistent_variable_get($prefix, $bin, $name, $default = NULL) {
  $data = variable_get($prefix . '_' . $name . '_' . $bin);
  if ($bin && !isset($data)) {
    $data = variable_get($prefix . '_default_' . $name, $default);
  }
  return isset($data) ? $data : $default;
}

/**
 * Defines a transactional aware cache implementation.
 *
 * Buffer/discard cache operations during transactions.
 */
class ConsistentCacheBase implements DrupalCacheInterface, DrupalTransactionalCacheInterface {
  /**
   * The cache bin for this instance.
   *
   * @var string
   */
  protected $bin;

  /**
   * The real cache backend to use for cache operations.
   *
   * @var DrupalCacheInterface
   */
  protected $backend = NULL;

  /**
   * The cache operations buffer.
   *
   * @var array
   *   Array operations, where each operation is an array containing:
   *     0: Transactional depth (0 - n).
   *     1: The operation to perform (set, clear).
   *     2: The cache ID.
   *     3: The data if "set", or wildcard boolean if "clear".
   *     4: The expiration of the cache entry if "set".
   */
  protected $buffer = array();

  /**
   * Safe mode. If TRUE, cache set operations are skipped during transactions.
   *
   * @var boolean
   */
  protected $safe = TRUE;

  /**
   * Strict mode. If TRUE, cache expiration is checked during fetch.
   *
   * @var boolean
   */
  protected $strict = FALSE;

  /**
   * Backend singletons per cache bin.
   *
   * @var array
   */
  static protected $backends = array();

  /**
   * The default cache backend to use.
   *
   * @var string
   */
  protected $defaultClass = 'DrupalDatabaseCache';

  /**
   * Constructor.
   *
   * Initialize bin.
   */
  public function __construct($bin) {
    $this->bin = $bin;

    // Let our parent do its work.
    $class = cache_consistent_variable_get('consistent_cache', $this->bin, 'class', $this->defaultClass);

    // Check for safe mode.
    $this->safe = cache_consistent_variable_get('consistent_cache', $this->bin, 'safe', TRUE);

    // Check for strict mode.
    $this->strict = cache_consistent_variable_get('consistent_cache', $this->bin, 'strict', FALSE);

    if (!isset(self::$backends[$bin])) {
      self::$backends[$bin] = new $class($bin);
    }
    $this->backend = &self::$backends[$bin];
  }

  /**
   * Destructor.
   *
   * Flush remaining buffers, if any.
   */
  public function __destruct() {
    $this->commit(0);
    unset(self::$backends[$this->bin]);
  }

  /**
   * Implements DrupalCacheInterface::get().
   */
  public function get($cid) {
    $cids = array($cid);
    $cache = $this->getMultiple($cids);
    return reset($cache);
  }

  /**
   * Find an entry in the buffer that matches the $cid.
   *
   * With or without wildcard.
   *
   * @param string $cid
   *   The cache ID to look up.
   *
   * @return array
   *   Cache operation if found, NULL otherwise.
   */
  protected function matchKey($cid) {
    if (!empty($this->buffer)) {
      $cid = (string) $cid;
      $operation = end($this->buffer);
      do {
        $key = (string) $operation[CC_CID];
        if ($cid === $key || strpos($cid, $key) === 0) {
          return $operation;
        }
      } while ($operation = prev($this->buffer));
    }
  }

  /**
   * Implements DrupalCacheInterface::getMultiple().
   *
   * Get multiple cache entries taking buffer into account.
   */
  public function getMultiple(&$cids) {
    $return = array();
    $fetch_cids = array_combine($cids, $cids);

    // If buffer is not empty, then check for changes in $cids requested.
    if (!empty($this->buffer)) {
      foreach ($cids as $cid) {
        if ($operation = $this->matchKey($cid)) {
          switch ($operation[CC_OPERATION]) {
            case 'set':
              $return[$cid] = (object) array(
                'cid' => $cid,
                'data' => $operation[CC_DATA_WILDCARD],
                'created' => time(),
                'headers' => NULL,
                'flushes' => 0,
                'expire' => $operation[CC_EXPIRE],
              );
              unset($fetch_cids[$cid]);
              break;

            case 'clear':
              unset($fetch_cids[$cid]);
              break;

          }
        }
      }
    }

    // Re-assign cache IDs found in buffer to $cids, to inform the consumer.
    $cids = array_combine(array_keys($return), array_keys($return));
    if ($fetch_cids) {
      // Fetch relevant cache entries from the backend, and add to $cids, while
      // maintaining unique cache IDs.
      $return += $this->backend->getMultiple($fetch_cids);
      $cids += array_combine($fetch_cids, $fetch_cids);
    }

    // Weed out expired cache entries.
    if ($this->strict) {
      foreach ($return as $cid => $cache) {
        if ($cache->expire > 0 && $cache->expire < time()) {
          unset($return[$cid]);
          $cids[] = $cid;
        }
      }
    }

    return $return;
  }

  /**
   * Implements DrupalCacheInterface::set().
   *
   * Buffer cache operation if inside transaction.
   */
  public function set($cid, $data, $expire = CACHE_PERMANENT) {
    $depth = drupal_bootstrap(NULL, FALSE) >= DRUPAL_BOOTSTRAP_DATABASE ? Database::getConnection()->transactionDepth() : 0;
    if ($depth == 0) {
      // Don't buffer untransactionalized cache operations.
      return $this->backend->set($cid, $data, $expire);
    }
    elseif (!$this->safe) {
      // Only buffer during transactions when safe mode is off.
      // Otherwise, discard cache set.
      $this->buffer[] = array(
        CC_DEPTH         => $depth,
        CC_OPERATION     => 'set',
        CC_CID           => $cid,
        CC_DATA_WILDCARD => $this->deepClone($data),
        CC_EXPIRE        => $expire,
      );
    }
  }

  /**
   * Implements DrupalCacheInterface::clear().
   *
   * Buffer cache operation if inside transaction.
   */
  public function clear($cid = NULL, $wildcard = FALSE) {
    // Explicit memcache support :-(
    if ($this->backend instanceOf MemCacheDrupal) {
      $backtrace = debug_backtrace();
      if ($cid == MEMCACHE_CONTENT_CLEAR || (isset($backtrace[2]) && $backtrace[2]['function'] == 'cache_clear_all' && empty($backtrace[2]['args']))) {
        $cid = MEMCACHE_CONTENT_CLEAR;
      }
    }

    $depth = drupal_bootstrap(NULL, FALSE) >= DRUPAL_BOOTSTRAP_DATABASE ? Database::getConnection()->transactionDepth() : 0;
    if ($depth == 0) {
      // Don't buffer untransactionalized cache operations.
      return $this->backend->clear($cid, $wildcard);
    }
    else {
      $cids = is_array($cid) ? $cid : array($cid);
      foreach ($cids as $cid) {
        $this->buffer[] = array(
          CC_DEPTH => $depth,
          CC_OPERATION => 'clear',
          CC_CID => $cid,
          CC_DATA_WILDCARD => $wildcard,
        );
      }
    }
  }

  /**
   * Implements DrupalCacheInterface::isEmpty().
   */
  public function isEmpty() {
    // If anything is present in the buffer, then we cannot determine if the
    // bin is empty.
    if (!empty($this->buffer)) {
      return FALSE;
    }

    // If inside a transaction in safe mode, then nothing is certain.
    if ($this->safe) {
      $depth = drupal_bootstrap(NULL, FALSE) >= DRUPAL_BOOTSTRAP_DATABASE ? Database::getConnection()->transactionDepth() : 0;
      if ($depth) {
        return FALSE;
      }
    }

    // Nothing in buffer nor safe+transaction; ask the backend.
    return $this->backend->isEmpty();
  }

  /**
   * Implements DrupalTransactionalCacheInterface::rollback().
   *
   * Clear cache operation buffer.
   *
   * Clear all affected cids from cache.
   * This is called only during rollback scenarios.
   */
  public function rollback($depth) {
    foreach ($this->buffer as $idx => $operation) {
      if ($operation[CC_DEPTH] > $depth) {
        unset($this->buffer[$idx]);
      }
    }
  }

  /**
   * Implements DrupalTransactionalCacheInterface::rollback().
   *
   * Flush cache operation buffer.
   */
  public function commit($depth) {
    if ($depth > 0) {
      // Nested commit. Don't flush yet, but move all operations to
      // parent depth (pseudo-commit).
      foreach ($this->buffer as $idx => &$operation) {
        if ($operation[CC_DEPTH] > $depth) {
          $operation[CC_DEPTH] = $depth;
        }
      }
      return;
    }

    // Flush the buffer, FIFO-style.
    reset($this->buffer);
    while (!empty($this->buffer)) {
      $operation = array_shift($this->buffer);
      $depth = array_shift($operation);
      $method = array_shift($operation);
      call_user_func_array(array($this->backend, $method), $operation);
    }
  }

  /**
   * Deep clone data structure.
   *
   * @param mixed $data
   *   The data structure to clone.
   *
   * @return mixed
   *   Cloned data structure.
   */
  protected function deepClone($data) {
    if (is_array($data)) {
      $result = array();
      foreach ($data as $key => $value) {
        $result[$key] = $this->deepClone($value);
      }
      return $result;
    }
    elseif (is_object($data)) {
      return clone $data;
    }
    else {
      return $data;
    }
  }

}


if (interface_exists('DrupalCacheDeferredInterface', FALSE)) {
  /**
   * Cache Consistent class with Cache Heuristic.
   */
  class ConsistentCache extends ConsistentCacheBase implements DrupalCacheDeferredInterface {
    /**
     * The default cache backend to use.
     *
     * When using Cache Heuristic, we want
     * to use the deferred database cache by default.
     *
     * @var string
     */
    protected $defaultClass = 'DrupalDatabaseCacheDeferred';

    /**
     * Implements DrupalCacheDeferredInterface::getMultipleDeferred().
     */
    public function getMultipleDeferred($cids) {
      if ($this->backend instanceof DrupalCacheDeferredInterface) {
        return $this->backend->getMultipleDeferred($cids);
      }
      else {
        return array();
      }
    }

    /**
     * Implements DrupalCacheDeferredInterface::prepareDeferredItem().
     */
    public function prepareDeferredItem($cid, $item) {
      $result = $this->backend->prepareDeferredItem($cid, $item);
      if (!empty($this->buffer)) {
        if ($operation = $this->matchKey($cid)) {
          switch ($operation[CC_OPERATION]) {
            case 'set':
              $result = (object) array(
                'cid' => $cid,
                'data' => $operation[CC_DATA_WILDCARD],
                'created' => time(),
                'headers' => NULL,
                'flushes' => 0,
                'expire' => $operation[CC_EXPIRE],
              );
              break;

            case 'clear':
              $result = NULL;
              break;
          }
        }
      }

      // Sanity check, and always weed out expired entries on deferred items.
      // @todo Why is expire checking always necessary here? Should we respect
      // strict-mode here as well?
      if ($result && $result->expire > 0 && $result->expire < time()) {
        $result = NULL;
      }

      return $result;
    }

    /**
     * Implements DrupalCacheDeferredInterface::fetchDeferredItems().
     */
    public function fetchDeferredItems($cids, $result) {
      if ($this->backend instanceof DrupalCacheDeferredInterface) {
        return $this->backend->fetchDeferredItems($cids, $result);
      }
      else {
        return array();
      }
    }

  }
}
else {
  /**
   * Cache Consistent class without Cache Heuristic support.
   */
  class ConsistentCache extends ConsistentCacheBase {}
}
